use std::sync::Arc;

use axum::{
    extract::{Query, State},
    response::{Html, IntoResponse},
    routing::get,
    Router,
};
use openidconnect::{
    core::{CoreAuthenticationFlow, CoreClient, CoreErrorResponseType, CoreProviderMetadata},
    reqwest::async_http_client,
    AccessTokenHash, AuthorizationCode, ClaimsVerificationError, ClientId, CsrfToken,
    DiscoveryError, IssuerUrl, Nonce, OAuth2TokenResponse, PkceCodeChallenge, RedirectUrl,
    RefreshToken, RequestTokenError, Scope, SigningError, StandardErrorResponse, TokenResponse,
};
use serde::Deserialize;
use thiserror::Error;
use tokio::sync::mpsc;

#[derive(Error, Debug)]
pub enum Error {
    #[error("url parse error: {:?}", 0)]
    UrlParse(#[from] openidconnect::url::ParseError),

    #[error("discovery error: {:?}", 0)]
    Discovery(#[from] DiscoveryError<openidconnect::reqwest::Error<reqwest::Error>>),

    #[error("request token error: {:?}", 0)]
    RequestToken(
        #[from]
        RequestTokenError<
            openidconnect::reqwest::Error<reqwest::Error>,
            StandardErrorResponse<CoreErrorResponseType>,
        >,
    ),

    #[error("claims verification error: {:?}", 0)]
    ClaimsVerification(#[from] ClaimsVerificationError),

    #[error("signing error: {:?}", 0)]
    Signing(#[from] SigningError),

    #[error("server did not return an id token")]
    NoIdToken,

    #[error("invalid access token")]
    InvalidAccessToken,

    #[error("no response received")]
    NoResponse,

    #[error("csrf mismatch")]
    CsrfMismatch,
}

#[derive(Debug, Deserialize)]
struct ResponseData {
    pub code: String,
    pub state: String,
}

pub(crate) async fn login(
    issuer: &str,
    client_id: &str,
    scopes: &[String],
    refresh_token: &mut Option<String>,
) -> Result<String, Error> {
    let provider_metadata = CoreProviderMetadata::discover_async(
        IssuerUrl::new(issuer.to_string())?,
        async_http_client,
    )
    .await?;

    // Create an OpenID Connect client by specifying the client ID, client secret, authorization URL
    // and token URL.
    let client = CoreClient::from_provider_metadata(
        provider_metadata,
        ClientId::new(client_id.to_string()),
        None,
    )
    // Set the URL the user will be redirected to after the authorization process.
    .set_redirect_uri(RedirectUrl::new("http://[::1]:8080".to_string())?);

    // Generate a PKCE challenge.
    let (pkce_challenge, pkce_verifier) = PkceCodeChallenge::new_random_sha256();

    if let Some(refresh_token) = refresh_token {
        if let Ok(token_response) = client
            .exchange_refresh_token(&RefreshToken::new(refresh_token.to_string()))
            .request_async(async_http_client)
            .await
        {
            eprintln!("authenticated with oidc provider");
            return Ok(token_response.access_token().secret().clone());
        }
    }

    // Generate the full authorization URL.
    let mut auth = client.authorize_url(
        CoreAuthenticationFlow::AuthorizationCode,
        CsrfToken::new_random,
        Nonce::new_random,
    );

    for scope in scopes {
        auth = auth.add_scope(Scope::new(scope.to_string()));
    }
    let (auth_url, csrf_token, nonce) = auth
        // Set the PKCE code challenge.
        .set_pkce_challenge(pkce_challenge)
        .url();
    open::that(auth_url.to_string()).unwrap();
    eprintln!("a browser should have been opened with the url {auth_url}. please login with your oidc provider.");

    let (fuse_tx, mut fuse_rx) = mpsc::channel::<ResponseData>(1);
    let app = Router::new()
        .route("/", get(handle_post))
        .with_state(Arc::new(fuse_tx));

    let server = axum::Server::bind(&"[::1]:8080".parse().unwrap()).serve(app.into_make_service());

    let data = tokio::select! {
        x = fuse_rx.recv() => {
            x
        }
        _ = server => {
            None
        }
    };

    let data = data.ok_or(Error::NoResponse)?;

    // match csrf_state

    if *csrf_token.secret() != data.state {
        return Err(Error::CsrfMismatch);
    }

    let token_response = client
        .exchange_code(AuthorizationCode::new(data.code))
        // Set the PKCE code verifier.
        .set_pkce_verifier(pkce_verifier)
        .request_async(async_http_client)
        .await?;

    // Extract the ID token claims after verifying its authenticity and nonce.
    let id_token = token_response.id_token().ok_or_else(|| Error::NoIdToken)?;
    let claims = id_token.claims(&client.id_token_verifier(), &nonce)?;

    // Verify the access token hash to ensure that the access token hasn't been substituted for
    // another user's.
    if let Some(expected_access_token_hash) = claims.access_token_hash() {
        let actual_access_token_hash =
            AccessTokenHash::from_token(token_response.access_token(), &id_token.signing_alg()?)?;
        if actual_access_token_hash != *expected_access_token_hash {
            return Err(Error::InvalidAccessToken);
        }
    }

    if let Some(new_refresh_token) = token_response.refresh_token() {
        *refresh_token = Some(new_refresh_token.secret().to_string());
    }

    eprintln!("authenticated with oidc provider");
    Ok(token_response.access_token().secret().clone())
}

async fn handle_post(
    State(fuse_tx): State<Arc<mpsc::Sender<ResponseData>>>,
    Query(data): Query<ResponseData>,
) -> impl IntoResponse {
    fuse_tx.clone().send(data).await;
    Html("<html><body>Die Anmeldung war erfolgreich. Du kannst dieses Fenster jetzt schließen.<script>window.close()</script></body></html>")
}
