use std::process::ExitCode;

use clap::Parser;
use reqwest::{Body, Url};
use serde::{Deserialize, Serialize};
use tokio::io::stdin;
use tokio_util::io::ReaderStream;

use crate::auth::login;

mod auth;

#[derive(Serialize, Deserialize, Debug)]
pub struct Config {
    pub refresh_token: Option<String>,
    pub binurl: String,
    pub issuer: String,
    pub client_id: String,
    pub scopes: Vec<String>,
}

#[derive(Debug, Parser)]
pub struct Args {
    #[arg(short, long)]
    content_type: Option<String>,

    #[arg(short, long)]
    ttl: Option<String>,
}

impl Default for Config {
    fn default() -> Self {
        Self {
            refresh_token: None,
            binurl: "https://bin.zettoit.eu".to_string(),
            issuer: "https://auth.zettoit.eu/realms/zettoit".to_string(),
            client_id: "binctl".to_string(),
            scopes: vec!["zettoit-bin".to_string()],
        }
    }
}

#[tokio::main]
async fn main() {
    let mut cfg: Config = confy::load("binctl", None).unwrap_or_default();

    let args = Args::parse();
    let access_token = login(
        &cfg.issuer,
        &cfg.client_id,
        cfg.scopes.as_slice(),
        &mut cfg.refresh_token,
    )
    .await
    .unwrap();
    let mut bin = create_bin(&cfg.binurl, &access_token).await.unwrap();
    eprintln!("created bin at {}. uploading...", bin);
    bin.set_query(args.ttl.map(|x| format!("ttl={}", x)).as_deref());

    upload_to_bin(
        bin.as_ref(),
        &args
            .content_type
            .unwrap_or("application/octet-stream".to_string()),
    )
    .await
    .unwrap();

    let _ = confy::store("binctl", None, cfg);
    bin.set_query(None);
    print!("{bin}");
}

async fn create_bin(binurl: &str, access_token: &str) -> Result<Url, reqwest::Error> {
    let client = reqwest::Client::new();

    Ok(client
        .get(binurl)
        .header("Authorization", format!("Bearer {}", access_token))
        .send()
        .await?
        .url()
        .to_owned())
}

async fn upload_to_bin(url: &str, content_type: &str) -> Result<(), reqwest::Error> {
    let client = reqwest::Client::new();

    if let Err(error) = client
        .post(url)
        .header("Content-Type", content_type)
        .body(Body::wrap_stream(ReaderStream::new(stdin())))
        .send()
        .await?
        .text()
        .await
    {
        eprintln!("{:?}", error);
        std::process::exit(exitcode::TEMPFAIL);
    }

    Ok(())
}
