use axum::{
    http::StatusCode,
    response::{IntoResponse, Response},
};
use log::{debug, error};

#[derive(Debug, thiserror::Error)]
pub enum Error {
    #[error("io error: {:?}", 0)]
    Io(#[from] std::io::Error),
    #[error("time error: {:?}", 0)]
    Time(#[from] std::time::SystemTimeError),
    #[error("metadata error: {:?}", 0)]
    MetadataDe(#[from] toml::de::Error),
    #[error("metadata error: {:?}", 0)]
    MetadataSer(#[from] toml::ser::Error),
    #[error("phrase is not valid")]
    PhraseInvalid,
    #[error("bin could not be found")]
    BinNotFound,
    #[error("file exists")]
    DataFileExists,

    #[error("hex error: {:?}", 0)]
    Hex(#[from] hex::FromHexError),

    #[error("oidc redirect")]
    Oidc(Response),

    #[error("invalid multipart")]
    InvalidMultipart,

    #[error("invalid ttl")]
    InvalidTtl,

    #[error("unauthorized")]
    Unauthorized,

    #[error("forbidden")]
    Forbidden,

    #[error("render error: {0:?}")]
    Template(#[from] sailfish::RenderError),

    #[error("prometheus: {0:?}")]
    Prometheus(std::fmt::Error),

    #[error("jwt invalid")]
    JwtInvalid,

    #[error("reqwest {0:?}")]
    Reqwest(#[from] reqwest::Error),

    #[error("reqwest {0:?}")]
    JwtDecode(#[from] jsonwebtoken::errors::Error),
}

impl IntoResponse for Error {
    fn into_response(self) -> axum::response::Response {
        debug!("{:?}", &self);
        match self {
            Self::PhraseInvalid => (StatusCode::BAD_REQUEST, "url is not valid\n").into_response(),
            Self::BinNotFound => (StatusCode::NOT_FOUND, "bin does not exist\n").into_response(),
            Self::DataFileExists => {
                (StatusCode::CONFLICT, "bin already contains data\n").into_response()
            }
            Self::Oidc(response) => response.into_response(),
            Self::InvalidMultipart => {
                (StatusCode::BAD_REQUEST, "invalid multipart data").into_response()
            }
            Self::InvalidTtl => (StatusCode::BAD_REQUEST, "invalid ttl specified").into_response(),
            Self::Unauthorized => (StatusCode::UNAUTHORIZED, "unauthorized\n").into_response(),
            Self::JwtInvalid => (StatusCode::BAD_REQUEST, "jwt token is invalid\n").into_response(),
            Self::Forbidden => (StatusCode::FORBIDDEN, "forbidden\n").into_response(),
            _ => {
                error!("{:?}", self);
                (StatusCode::INTERNAL_SERVER_ERROR, "internal server error\n").into_response()
            }
        }
        .into_response()
    }
}
