use std::{
    collections::BinaryHeap,
    sync::Arc,
    time::{Duration, SystemTime, UNIX_EPOCH},
};

use log::{debug, info, warn};
use tokio::{fs, sync::Mutex};

use crate::{metadata::Metadata, util::Id, AppMetrics, BinDownloadLabels};

#[derive(Clone)]
pub(crate) struct GarbageCollector {
    heap: Arc<Mutex<BinaryHeap<GarbageCollectorItem>>>,
    path: String,
    metrics: Arc<AppMetrics>,
}

impl GarbageCollector {
    pub(crate) async fn new(path: String, metrics: Arc<AppMetrics>) -> Self {
        let mut heap = BinaryHeap::new();
        let mut dir = fs::read_dir(&path).await.expect("readable data dir");
        while let Ok(Some(entry)) = dir.next_entry().await {
            let file_name = entry.file_name();
            let file_name = file_name.to_string_lossy();
            if let Some(id) = file_name.strip_suffix(".toml") {
                if let Ok(metadata) = Metadata::from_file(&format!("{}/{}.toml", path, id)).await {
                    metrics.bin_total.inc();
                    metrics
                        .bin_data_size
                        .inc_by(metadata.size.map(|x| x as i64).unwrap_or_default());
                    let _ = metrics.bin_downloads.get_or_create(&BinDownloadLabels {
                        bin: Id::from_str(id),
                    });
                    heap.push(GarbageCollectorItem {
                        expires_at: metadata.expires_at,
                        id: Id::from_str(id),
                    })
                }
            }
        }
        Self {
            heap: Arc::new(Mutex::new(heap)),
            path,
            metrics,
        }
    }

    pub(crate) async fn schedule(&self, id: &Id, expires_at: u64) {
        self.heap.lock().await.push(GarbageCollectorItem {
            id: id.clone(),
            expires_at,
        })
    }

    pub(crate) async fn run(&self) {
        let mut heap = self.heap.lock().await;

        let now = SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .expect("time after EPOCH")
            .as_secs();

        // skip if first item is not old enough to be deleted
        if let Some(true) = heap.peek().map(|x| x.expires_at > now) {
            return;
        }
        if let Some(item) = heap.pop() {
            if let Ok(metadata) =
                Metadata::from_file(&format!("{}/{}.toml", self.path, item.id)).await
            {
                // check metadata if the item is really ready for deletion (needed for reschedule
                // population of bin with data)
                if metadata.expires_at <= now {
                    debug!("deleting bin {}", item.id);
                    let res_meta =
                        fs::remove_file(&format!("{}/{}.toml", self.path, item.id)).await;
                    let res_data = fs::remove_file(&format!("{}/{}.dat", self.path, item.id)).await;

                    if res_meta.is_err() || res_data.is_err() {
                        warn!("failed to delete bin {} for gc", item.id);
                    }

                    self.metrics.bin_total.dec();
                    self.metrics
                        .bin_data_size
                        .dec_by(metadata.size.map(|x| x as i64).unwrap_or_default());
                }
            } else {
                info!("cant open metadata file for bin {} for gc", item.id);
            }
        }
    }

    pub(crate) fn spawn(&self) {
        let gc = self.clone();
        tokio::spawn(async move {
            loop {
                tokio::time::sleep(Duration::from_secs(10)).await;
                gc.run().await;
            }
        });
    }
}

#[derive(PartialEq, Eq)]
struct GarbageCollectorItem {
    id: Id,
    expires_at: u64,
}

impl PartialOrd for GarbageCollectorItem {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for GarbageCollectorItem {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        self.expires_at.cmp(&other.expires_at).reverse()
    }
}
