use std::marker::PhantomData;

use axum::{
    async_trait,
    extract::{FromRef, FromRequestParts},
    http::request::Parts,
};
use jsonwebtoken::{decode, Algorithm, DecodingKey, Validation};
use serde::{de::DeserializeOwned, Deserialize};

use crate::error::Error;

#[derive(Debug, Clone, Deserialize)]
pub struct Claims<A: Clone> {
    pub aud: Vec<String>,
    pub exp: usize,
    pub iat: usize,
    pub iss: String,
    pub sub: String,
    pub azp: String,

    pub name: String,
    pub preferred_username: String,
    pub given_name: String,
    pub family_name: String,
    pub email: String,
    #[serde(flatten)]
    pub additional: A,
}

#[derive(Clone)]
pub struct JwtApplication<A: Clone> {
    validation: Validation,
    pubkey: DecodingKey,
    _a: PhantomData<A>,
}

#[derive(Deserialize)]
struct IssuerDiscovery {
    public_key: String,
}

impl<A: Clone + DeserializeOwned> JwtApplication<A> {
    pub async fn new(issuer: String, audience: String) -> Result<Self, Error> {
        let issuer_key = reqwest::get(&issuer)
            .await?
            .json::<IssuerDiscovery>()
            .await?
            .public_key;

        let pem = format!(
            "-----BEGIN PUBLIC KEY-----\n{}\n-----END PUBLIC KEY-----",
            issuer_key
        );

        let pubkey = DecodingKey::from_rsa_pem(pem.as_bytes())?;

        let mut validation = Validation::new(Algorithm::RS256);
        validation.set_issuer(&[issuer]);
        validation.set_audience(&[audience]);
        validation.validate_nbf = true;

        Ok(Self {
            validation,
            pubkey,
            _a: PhantomData,
        })
    }
}

#[async_trait]
impl<S, A> FromRequestParts<S> for Claims<A>
where
    S: Send + Sync,
    A: Clone + DeserializeOwned,
    JwtApplication<A>: FromRef<S>,
{
    type Rejection = Error;

    async fn from_request_parts(parts: &mut Parts, state: &S) -> Result<Self, Self::Rejection> {
        let application: JwtApplication<A> = JwtApplication::from_ref(state);

        let token = parts
            .headers
            .get("Authorization")
            .and_then(|x| x.to_str().ok())
            .map(|x| x.chars().skip(7).collect::<String>());

        let token = token.and_then(|x| {
            decode::<Claims<A>>(&x, &application.pubkey, &application.validation).ok()
        });

        if let Some(token) = token {
            Ok(token.claims)
        } else {
            Err(Error::JwtInvalid)
        }
    }
}
