#![deny(clippy::unwrap_used)]
use axum_oidc::{
    error::{ExtractorError, MiddlewareError},
    EmptyAdditionalClaims, OidcAuthLayer, OidcClaims, OidcLoginLayer,
};
use duration_str::{deserialize_option_duration, parse_std};
use jwt::JwtApplication;
use prometheus_client::{
    encoding::EncodeLabelSet,
    metrics::{counter::Counter, family::Family, gauge::Gauge},
    registry::{Registry, Unit},
};
use sailfish::TemplateOnce;
use std::{
    borrow::Borrow,
    env,
    str::FromStr,
    sync::Arc,
    time::{Duration, SystemTime, UNIX_EPOCH},
};
use tower_http::services::ServeDir;

use tower::ServiceBuilder;
use tower_sessions::{cookie::SameSite, MemoryStore, SessionManagerLayer};

use axum::{
    async_trait,
    body::Body,
    error_handling::HandleErrorLayer,
    extract::{FromRef, FromRequest, Multipart, Path, Query, State},
    http::{
        header::{self, CONTENT_TYPE},
        HeaderMap, HeaderValue, Request, StatusCode, Uri,
    },
    response::{Html, IntoResponse, Redirect, Response},
    routing::{delete, get, post},
    BoxError, Router,
};

use chacha20::{
    cipher::{KeyIvInit, StreamCipher},
    ChaCha20,
};
use futures_util::StreamExt;
use garbage_collector::GarbageCollector;
use log::{debug, warn};
use serde::Deserialize;
use sha3::{Digest, Sha3_256};
use tokio::{
    fs::{self, File},
    io::{AsyncWriteExt, BufReader, BufWriter},
    net::TcpListener,
};
use util::{IdSalt, KeySalt};

use crate::{
    error::Error,
    jwt::Claims,
    metadata::Metadata,
    util::{Id, Key, Nonce, Phrase},
    web_util::DecryptingStream,
};
mod error;
mod garbage_collector;
mod jwt;
mod metadata;
mod util;
mod web_util;

/// length of the "phrase" that is used to access the bin (https://example.com/<phrase>)
const PHRASE_LENGTH: usize = 16;
/// length of the salts that are used to generate the key for chacha and id for the files
const SALT_LENGTH: usize = 256 - PHRASE_LENGTH;

type HandlerResult<T> = Result<T, Error>;
#[derive(Clone)]
pub struct AppState {
    application_base: String,
    data: String,
    jwt_application: JwtApplication<EmptyAdditionalClaims>,
    key_salt: KeySalt,
    id_salt: IdSalt,
    garbage_collector: GarbageCollector,
    prometheus_registry: Arc<Registry>,
    metrics: Arc<AppMetrics>,
}

#[derive(Clone, Default)]
pub struct AppMetrics {
    bin_total: Gauge,
    bin_data_size: Gauge,
    bin_downloads: Family<BinDownloadLabels, Counter>,
}

#[derive(EncodeLabelSet, PartialEq, Eq, Hash, Clone, Debug)]
pub struct BinDownloadLabels {
    bin: Id,
}

impl FromRef<AppState> for JwtApplication<EmptyAdditionalClaims> {
    fn from_ref(input: &AppState) -> Self {
        input.jwt_application.clone()
    }
}

#[derive(PartialEq, Eq, PartialOrd, Ord, Debug)]
pub struct ExpiryHeapItem {
    pub expires_at: u64,
    pub id: String,
}

#[tokio::main]
async fn main() {
    dotenvy::dotenv().ok();
    env_logger::init();

    let application_base = env::var("APPLICATION_BASE").expect("APPLICATION_BASE env var");
    let issuer = env::var("ISSUER").expect("ISSUER env var");
    let client_id = env::var("CLIENT_ID").expect("CLIENT_ID env var");
    let client_secret = env::var("CLIENT_SECRET").ok();
    let scopes = env::var("SCOPES")
        .expect("SCOPES env var")
        .split(' ')
        .map(|x| x.to_owned())
        .collect::<Vec<_>>();

    let session_store = MemoryStore::default();
    let session_service = SessionManagerLayer::new(session_store).with_same_site(SameSite::Lax);

    let oidc_login_service = ServiceBuilder::new()
        .layer(HandleErrorLayer::new(|e: MiddlewareError| async {
            dbg!("{:?}", &e);
            e.into_response()
        }))
        .layer(OidcLoginLayer::<EmptyAdditionalClaims>::new());

    let oidc_auth_service = ServiceBuilder::new()
        .layer(HandleErrorLayer::new(|e: MiddlewareError| async {
            dbg!("{:?}", &e);
            e.into_response()
        }))
        .layer(
            OidcAuthLayer::<EmptyAdditionalClaims>::discover_client(
                Uri::from_maybe_shared(application_base.clone()).expect("valid base url"),
                issuer.clone(),
                client_id,
                client_secret,
                scopes,
            )
            .await
            .expect("Oidc Layer"),
        );

    let jwt_application =
        JwtApplication::new(issuer, env::var("AUDIENCE").expect("AUDIENCE env var"))
            .await
            .expect("Jwt Authentication Client");

    let data_path = env::var("DATA_PATH").expect("DATA_PATH env var");

    let mut registry = Registry::default();

    let app_metrics = Arc::new(AppMetrics::default());

    registry.register(
        "bin_total",
        "total number of stored bins",
        app_metrics.bin_total.clone(),
    );

    registry.register(
        "bin_downloads",
        "number of times bins got accessed",
        app_metrics.bin_downloads.clone(),
    );

    registry.register_with_unit(
        "bin_data_size",
        "combined size of all bins",
        Unit::Bytes,
        app_metrics.bin_data_size.clone(),
    );

    let garbage_collector = GarbageCollector::new(data_path.clone(), app_metrics.clone()).await;
    garbage_collector.spawn();

    let state: AppState = AppState {
        application_base,
        data: data_path,
        jwt_application,
        key_salt: KeySalt::from_str(&env::var("KEY_SALT").expect("KEY_SALT env var"))
            .expect("KEY SALT valid hex"),
        id_salt: IdSalt::from_str(&env::var("ID_SALT").expect("ID_SALT env var"))
            .expect("ID_SALT valid hex"),
        garbage_collector,
        prometheus_registry: Arc::new(registry),
        metrics: app_metrics,
    };

    // when the two salts are identical, the bin id and the bin key are also identical, this would
    // make the encryption useless
    assert_ne!(state.key_salt.raw(), state.id_salt.raw());

    let app = Router::new()
        .route("/", get(get_index))
        .route("/:id", delete(delete_bin))
        .route("/:id/delete", get(delete_bin_interactive).post(delete_bin))
        .layer(oidc_login_service)
        .route("/:id", get(get_item).post(upload_bin).put(upload_bin))
        .route("/metrics", get(metrics))
        .nest_service("/static", ServeDir::new("static"))
        .with_state(state)
        .layer(oidc_auth_service)
        .layer(session_service);

    let listener = TcpListener::bind("[::]:8080")
        .await
        .expect("valid listen address");
    axum::serve(listener, app).await.expect("axum server");
}

async fn get_index(
    State(app_state): State<AppState>,
    oidc_extractor: Result<OidcClaims<EmptyAdditionalClaims>, ExtractorError>,
    jwt_claims: Option<Claims<EmptyAdditionalClaims>>,
) -> HandlerResult<impl IntoResponse> {
    let subject = match (oidc_extractor, jwt_claims) {
        (_, Some(claims)) => claims.sub.to_string(),
        (Ok(oidc), None) => oidc.0.subject().to_string(),
        (Err(err), None) => return Err(Error::Oidc(err.into_response())),
    };

    //generate phrase and derive id from it
    let phrase = Phrase::random();
    let id = Id::from_phrase(&phrase, &app_state.id_salt);

    let nonce = Nonce::random();

    let expires_at = SystemTime::now().duration_since(UNIX_EPOCH)?.as_secs() + 24 * 3600;

    let metadata_path = format!("{}/{}.toml", app_state.data, id);
    let metadata = Metadata {
        subject,
        nonce: nonce.to_hex(),
        etag: None,
        size: None,
        content_type: None,
        expires_at,
    };
    metadata.to_file(&metadata_path).await?;

    app_state.garbage_collector.schedule(&id, expires_at).await;

    debug!("created bin {id}");

    Ok(Redirect::temporary(&format!(
        "{}{}",
        app_state.application_base, phrase
    )))
}

async fn delete_bin(
    Path(phrase): Path<String>,
    State(app_state): State<AppState>,
    oidc_extractor: Result<OidcClaims<EmptyAdditionalClaims>, ExtractorError>,
    jwt_claims: Option<Claims<EmptyAdditionalClaims>>,
) -> HandlerResult<impl IntoResponse> {
    let subject = match (oidc_extractor, jwt_claims) {
        (_, Some(claims)) => claims.sub.to_string(),
        (Ok(oidc), None) => oidc.0.subject().to_string(),
        (Err(_), None) => return Err(Error::Unauthorized),
    };

    let phrase = Phrase::from_str(&phrase)?;
    let id = Id::from_phrase(&phrase, &app_state.id_salt);

    let metadata_path = format!("{}/{}.toml", app_state.data, id);
    let metadata = Metadata::from_file(&metadata_path).await?;

    if metadata.subject != subject {
        return Err(Error::Forbidden);
    }

    debug!("deleting bin {}", id);
    let res_meta = fs::remove_file(&format!("{}/{}.toml", app_state.data, id)).await;
    let res_data = fs::remove_file(&format!("{}/{}.dat", app_state.data, id)).await;

    if res_meta.is_err() || res_data.is_err() {
        warn!("failed to delete bin {} for manual deletion", id);
    }

    app_state.metrics.bin_total.dec();
    app_state
        .metrics
        .bin_data_size
        .dec_by(metadata.size.map(|x| x as i64).unwrap_or_default());

    Ok("ok\n")
}

async fn delete_bin_interactive(
    _: Path<String>,
    _: OidcClaims<EmptyAdditionalClaims>,
) -> HandlerResult<impl IntoResponse> {
    Ok(Html(DeleteTemplate.render_once()?))
}

#[derive(Deserialize)]
pub struct PostQuery {
    ttl: Option<String>,
}

async fn upload_bin(
    Path(phrase): Path<String>,
    Query(params): Query<PostQuery>,
    State(app_state): State<AppState>,
    headers: HeaderMap,
    data: MultipartOrStream,
) -> HandlerResult<impl IntoResponse> {
    let phrase = Phrase::from_str(&phrase)?;
    let id = Id::from_phrase(&phrase, &app_state.id_salt);

    let metadata_path = format!("{}/{}.toml", app_state.data, id);
    let mut metadata = Metadata::from_file(&metadata_path).await?;

    let path = format!("{}/{}.dat", app_state.data, id);
    let path = std::path::Path::new(&path);

    if !path.exists() {
        let key = Key::from_phrase(&phrase, &app_state.key_salt);
        let nonce = Nonce::from_hex(&metadata.nonce)?;
        let mut cipher = ChaCha20::new(key.borrow(), nonce.borrow());

        let file = File::create(&path).await?;
        let mut writer = BufWriter::new(file);

        let mut etag_hasher = Sha3_256::new();
        let mut size: u64 = 0;

        let mut ttl = params
            .ttl
            .map(|x| duration_str::parse(&x).map_err(|_| Error::InvalidTtl))
            .transpose()?
            .unwrap_or(Duration::from_secs(24 * 3600));

        match data {
            MultipartOrStream::Stream(stream) => {
                let mut stream = stream.into_data_stream();
                while let Some(chunk) = stream.next().await {
                    let mut buf = chunk.unwrap_or_default().to_vec();
                    etag_hasher.update(&buf);
                    size += buf.len() as u64;
                    cipher.apply_keystream(&mut buf);
                    writer.write_all(&buf).await?;
                }
                metadata.content_type = match headers.get(CONTENT_TYPE) {
                    Some(content_type) => Some(
                        content_type
                            .to_owned()
                            .to_str()
                            .unwrap_or_default()
                            .to_string(),
                    ),
                    None => Some("application/octet-stream".to_string()),
                };
            }
            MultipartOrStream::Multipart(mut multipart) => {
                let mut file_read = false;
                while let Some(mut field) = multipart
                    .next_field()
                    .await
                    .map_err(|_| Error::InvalidMultipart)?
                {
                    if field.name().unwrap_or_default() == "file" && !file_read {
                        while let Some(chunk) = field.chunk().await.unwrap_or_default() {
                            let mut buf = chunk.to_vec();
                            etag_hasher.update(&buf);
                            size += buf.len() as u64;
                            cipher.apply_keystream(&mut buf);
                            writer.write_all(&buf).await?;
                        }
                        metadata.content_type = Some(
                            field
                                .content_type()
                                .map(|x| x.to_string())
                                .unwrap_or("application/octet-stream".to_string()),
                        );
                        file_read = true;
                    }
                    if field.name().unwrap_or_default() == "ttl" {
                        if let Some(mp_ttl) =
                            field.text().await.ok().and_then(|x| parse_std(x).ok())
                        {
                            ttl = mp_ttl;
                        }
                    }
                }
            }
        }
        writer.flush().await?;

        if let Some(expires_at) = SystemTime::now()
            .duration_since(UNIX_EPOCH)?
            .checked_add(ttl)
            .map(|x| x.as_secs())
        {
            metadata.expires_at = expires_at;
        }

        metadata.etag = Some(hex::encode(etag_hasher.finalize()));
        metadata.size = Some(size);

        metadata.to_file(&metadata_path).await?;

        app_state
            .garbage_collector
            .schedule(&id, metadata.expires_at)
            .await;

        debug!("bin {id} got filled");

        app_state.metrics.bin_total.inc();
        app_state
            .metrics
            .bin_data_size
            .inc_by(metadata.size.map(|x| x as i64).unwrap_or_default());

        Ok((StatusCode::OK, "ok\n"))
    } else {
        Err(Error::DataFileExists)
    }
}

async fn get_item(
    Path(phrase): Path<String>,
    State(app_state): State<AppState>,
) -> HandlerResult<impl IntoResponse> {
    let phrase = Phrase::from_str(&phrase)?;
    let id = Id::from_phrase(&phrase, &app_state.id_salt);

    let metadata = Metadata::from_file(&format!("{}/{}.toml", app_state.data, id)).await?;

    let path = format!("{}/{}.dat", app_state.data, id);

    let key = Key::from_phrase(&phrase, &app_state.key_salt);
    let nonce = Nonce::from_hex(&metadata.nonce)?;

    if std::fs::metadata(&path).is_err() {
        Ok((
            StatusCode::ACCEPTED,
            Html(
                IndexTemplate {
                    bin_url: &format!("{}{}", app_state.application_base, phrase),
                }
                .render_once()?,
            ),
        )
            .into_response())
    } else {
        //TODO(pfz4): Maybe add link handling
        let file = File::open(&path).await?;
        let reader = BufReader::new(file);

        let body = Body::from_stream(DecryptingStream::new(
            reader,
            id.clone(),
            &metadata,
            &key,
            &nonce,
        ));

        let mut headers = HeaderMap::new();
        headers.insert(
            header::CONTENT_LENGTH,
            metadata.size.unwrap_or_default().into(),
        );
        if let Ok(subject) = metadata.subject.parse() {
            headers.insert("CreatedBy", subject);
        }

        if let Some(content_type) = metadata.content_type.and_then(|x| x.parse().ok()) {
            headers.insert(header::CONTENT_TYPE, content_type);
        }
        if let Some(etag) = metadata.etag.clone().and_then(|x| x.parse().ok()) {
            headers.insert(header::ETAG, etag);
        }
        if let Some(digest) = metadata
            .etag
            .and_then(|x| format!("sha3-256={x}").parse().ok())
        {
            headers.insert("Digest", digest);
        }

        app_state
            .metrics
            .bin_downloads
            .get_or_create(&BinDownloadLabels { bin: id })
            .inc();

        Ok((StatusCode::OK, headers, body).into_response())
    }
}

async fn metrics(State(app_state): State<AppState>) -> HandlerResult<impl IntoResponse> {
    let mut buffer = String::new();
    prometheus_client::encoding::text::encode(&mut buffer, &app_state.prometheus_registry)
        .map_err(Error::Prometheus)?;

    let mut headers = HeaderMap::new();
    headers.insert(
        CONTENT_TYPE,
        HeaderValue::from_static("text/plain; version=0.0.4"),
    );

    Ok((headers, buffer))
}

enum MultipartOrStream {
    Multipart(Multipart),
    Stream(Body),
}

#[async_trait]
impl<S: Sync + Send> FromRequest<S> for MultipartOrStream {
    type Rejection = Response;

    async fn from_request(req: Request<Body>, state: &S) -> Result<Self, Self::Rejection> {
        let is_multipart = req
            .headers()
            .get(CONTENT_TYPE)
            .and_then(|x| {
                x.to_str()
                    .ok()
                    .map(|y| y.starts_with("multipart/form-data"))
            })
            .unwrap_or_default();
        if is_multipart {
            Ok(Self::Multipart(
                Multipart::from_request(req, state)
                    .await
                    .map_err(|x| x.into_response())?,
            ))
        } else {
            Ok(Self::Stream(
                Body::from_request(req, state)
                    .await
                    .map_err(|x| x.into_response())?,
            ))
        }
    }
}

#[derive(TemplateOnce)]
#[template(path = "index.stpl")]
pub struct IndexTemplate<'a> {
    bin_url: &'a str,
}

#[derive(TemplateOnce)]
#[template(path = "delete.stpl")]
pub struct DeleteTemplate;
